@@ -1,11 +1,8 @@
-from . import aggregation_cm
-from . import amalgamation_cm
-
 try:
     from .aggregation_cm import aggregate
-    from .amalgamation_cm import amalgamate
+    from .merging_cm import fast_agglomerate as agglomerate 
 except ModuleNotFoundError:
     from .aggregation import aggregate
-    from .amalgamation import amalgamate
+    from .merging import fast_agglomerate as agglomerate 
     print("Cython fail.")
 from .clustering import CLASSIX
\ No newline at end of file
