@@ -1,37 +1,28 @@
 # -*- coding: utf-8 -*-
-
+#
 # CLASSIX: Fast and explainable clustering based on sorting
 #
-# Copyright (c) 2021, Stefan Güttel, Xinye Chen 
-# All rights reserved.
+# MIT License
 #
-# Redistribution and use in source and binary forms, with or without
-# modification, are permitted provided that the following conditions are met:
-# 1. Redistributions of source code must retain the above copyright notice, this
-#    list of conditions and the following disclaimer.
-# 2. Redistributions in binary form must reproduce the above copyright notice,
-#    this list of conditions and the following disclaimer in the documentation
-#    and/or other materials provided with the distribution.
-# 3. Neither the name of the copyright holder nor the names of its
-#    contributors may be used to endorse or promote products derived from
-#    this software without specific prior written permission.
+# Copyright (c) 2022 Stefan Güttel, Xinye Chen
 #
-# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
-# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
-# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
-# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
-# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
-# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-# 
-# Contacts: 
-#   Stefan Güttel : <stefan.guettel@manchester.ac.uk>  orcidID{0000-0003-1494-4478}
-#   Xinye Chen : <xinye.chen@manchester.ac.uk> orcidID{0000-0003-1778-393X}
+# Permission is hereby granted, free of charge, to any person obtaining a copy
+# of this software and associated documentation files (the "Software"), to deal
+# in the Software without restriction, including without limitation the rights
+# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+# copies of the Software, and to permit persons to whom the Software is
+# furnished to do so, subject to the following conditions:
 #
-# License: BSD 3 clause
+# The above copyright notice and this permission notice shall be included in all
+# copies or substantial portions of the Software.
+#
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+# SOFTWARE.
 
 try:
     # # %load_ext Cython
@@ -728,8 +719,9 @@ class CLASSIX:
     
     def explain(self, index1="NULL", index2="NULL", showsplist=True, max_colwidth=None, replace_name=None, 
                 plot=False, figsize=(10, 6), figstyle="ggplot", savefig=False, ind_color="k", ind_marker_size=150,
-                st_fcolor='tomato',  st_alpha=0.2, st_pad=2, bbox=None,
-                color='red', alpha=0.5, cline_width=0.5, figname=None, fmt='pdf'):
+                sp_fcolor='tomato',  sp_alpha=0.3, sp_pad=2, sp_fontsize=None, sp_bbox=None, 
+                dp_fcolor='bisque',  dp_alpha=0.6, dp_pad=2, dp_fontsize=None, dp_bbox=None,
+                color='red', connect_color='green', alpha=0.5, cline_width=0.5, axis='off', figname=None, fmt='pdf'):
         """
         'self.explain(object/index) # prints an explanation for why a point object1 is in its cluster (or an outlier)
         'self.explain(object1/index1, object2index) # prints an explanation why object1 and object2 are either in the same or distinct clusters
@@ -788,8 +780,8 @@ class CLASSIX:
         ind_marker_size : float), optional:
             Size for visualization of data with indices index1 and index2.
     
-        bbox : dict
-            Dict with properties for patches.FancyBboxPatch. Default as {'facecolor': 'tomato', 'alpha': 0.2, 'pad': 2}.
+        sp_bbox : dict
+            Dict with properties for patches.FancyBboxPatch for starting points. Default as {'facecolor': 'tomato', 'alpha': 0.2, 'pad': 2}.
         
         color : str
             Color for text of starting points labels in visualization. Default as "k".
@@ -808,7 +800,7 @@ class CLASSIX:
         
         """
         # deprecated (24/07/2021)
-        # cols = ["NrPts"] # 
+        # cols = ["NrPts"] 
         # for i in range(self.splist.shape[1] - 2):
         #     cols = cols + ["feat." + str(i + 1)]
         
@@ -822,12 +814,24 @@ class CLASSIX:
         #     spdisplay = pd.DataFrame(self.splist[:, 1:], columns=cols)
         
         # -----------------------------second method--------------------------------
-        if bbox is None:
-            bbox = dict()
-            bbox['facecolor'] = st_fcolor
-            bbox['alpha'] = st_alpha
-            bbox['pad'] = st_pad
-            # {'facecolor': 'tomato', 'alpha': 0.2, 'pad': 2},
+        if sp_bbox is None:
+            sp_bbox = dict()
+            sp_bbox['facecolor'] = sp_fcolor
+            sp_bbox['alpha'] = sp_alpha
+            sp_bbox['pad'] = sp_pad
+        # else: sp_bbox = {'facecolor': 'tomato', 'alpha': 0.2, 'pad': 2}
+        
+        if dp_bbox is None:
+            dp_bbox = dict()
+            dp_bbox['facecolor'] = dp_fcolor
+            dp_bbox['alpha'] = dp_alpha
+            dp_bbox['pad'] = dp_pad
+        # else: dp_bbox = {'facecolor': 'bisque', 'alpha': 0.2, 'pad': 2}
+            
+        if dp_fontsize is None and sp_fontsize is not None:
+            dp_fontsize = sp_fontsize
+            
+            
         if self.cluster_color is None:
             self.cluster_color = dict()
             for i in np.unique(self.labels_):
@@ -861,7 +865,7 @@ class CLASSIX:
             
         if index1 == "NULL": # analyze in the general way with a global view
             if plot == True:
-                self.explain_viz(figsize=figsize, figstyle=figstyle, savefig=savefig, bbox=bbox, fmt=fmt)
+                self.explain_viz(figsize=figsize, figstyle=figstyle, savefig=savefig, fontsize=sp_fontsize, bbox=sp_bbox, axis=axis, fmt=fmt)
                 
             data_size = self.data.shape[0]
             feat_dim = self.data.shape[1]
@@ -889,8 +893,10 @@ class CLASSIX:
         else: # explain(index1)
             if isinstance(index1, int):
                 object1 = self.x_pca[index1] # self.data has been normalized
+                agg_label1 = self.agg_labels[index1] # get the group index for object1
             else:
                 object1 = (index1 - self._mu) / self._scl # allow for out-sample data
+                agg_label1 = np.argmin(np.linalg.norm(self.s_pca - object1, axis=1, ord=2)) # get the group index for object1
                 if self.data.shape[1] >= 2:
                     object1 = self.pca.transform(object1)
                     
@@ -911,7 +917,6 @@ class CLASSIX:
                 else:
                     index1_name = index1
 
-                agg_label1 = np.argmin(np.linalg.norm(self.s_pca - object1, axis=1, ord=2))
                 cluster_label1 = self.label_change[agg_label1]
                 sp_str = self.splist[agg_label1, 3:]
                 
@@ -923,20 +928,32 @@ class CLASSIX:
                     x_pca = self.x_pca[self.labels_ == cluster_label1]
                     s_pca = self.s_pca[self.sp_info.Cluster == cluster_label1]
                     
-                    ax.scatter(x_pca[:, 0], x_pca[:, 1], marker="*", c=self.cluster_color[cluster_label1])
+                    ax.scatter(x_pca[:, 0], x_pca[:, 1], marker=".", c=self.cluster_color[cluster_label1])
                     # ax.scatter(x_pca[:, 0], x_pca[:, 1], marker="*", c='pink')
                     ax.scatter(s_pca[:, 0], s_pca[:, 1], marker="p")
-
+                    
                     ax.scatter(object1[0], object1[1], marker="*", s=ind_marker_size)
-                    ax.text(object1[0], object1[1], s=str(index1_name), color=ind_color)
-
+                        
+                    if dp_fontsize is None:
+                        ax.text(object1[0], object1[1], s=str(index1_name), bbox=dp_bbox, color=ind_color)
+                    else:
+                        ax.text(object1[0], object1[1], s=str(index1_name), fontsize=dp_fontsize, bbox=dp_bbox, color=ind_color)
+                        
                     for i in range(s_pca.shape[0]):
                         ax.add_patch(plt.Circle((s_pca[i, 0], s_pca[i, 1]), self.radius, fill=False, color=color, alpha=alpha, lw=cline_width, clip_on=False))
                         ax.set_aspect('equal', adjustable='datalim')
-                        ax.text(s_pca[i, 0], s_pca[i, 1],
-                                 s=str(self.sp_info.Group[self.sp_info.Cluster == cluster_label1].astype(int).values[i]),
-                                # self.splist[self.sp_info.Cluster == cluster_label1, 1][i].astype(int).astype(str), 
-                                 bbox=bbox)
+                        if sp_fontsize is None:
+                            ax.text(s_pca[i, 0], s_pca[i, 1],
+                                    s=str(self.sp_info.Group[self.sp_info.Cluster == cluster_label1].astype(int).values[i]),
+                                    # self.splist[self.sp_info.Cluster == cluster_label1, 1][i].astype(int).astype(str), 
+                                    bbox=sp_bbox
+                            )
+                        else:
+                            ax.text(s_pca[i, 0], s_pca[i, 1],
+                                    s=str(self.sp_info.Group[self.sp_info.Cluster == cluster_label1].astype(int).values[i]),
+                                    # self.splist[self.sp_info.Cluster == cluster_label1, 1][i].astype(int).astype(str), 
+                                    fontsize=sp_fontsize, bbox=sp_bbox
+                            )
                     ax.plot()
                     ax.axis('off') # the axis here may not be consistent, so hide.
                     if savefig:
@@ -971,7 +988,7 @@ class CLASSIX:
                     #         ax.set_aspect('equal', adjustable='datalim')
                     #         ax.text(s_pca[i, 0], s_pca[i, 1],
                     #                  s=self.splist[self.sp_info.Cluster == cluster_label1,1][i].astype(int).astype(str), 
-                    #                  bbox=bbox)
+                    #                  bbox=sp_bbox)
                     #     ax.plot()
                     #     if savefig:
                     #         if not os.path.exists("img"):
@@ -1000,7 +1017,7 @@ class CLASSIX:
                     #         ax.set_aspect('equal', adjustable='datalim')
                     #         plt.text(s_pca[i, 0], s_pca[i, 1],
                     #                  s=self.splist[self.sp_info.Cluster == cluster_label1,1][i].astype(int).astype(str), 
-                    #                  bbox=bbox)
+                    #                  bbox=sp_bbox)
                     #     ax.plot() 
                     #     if savefig:
                     #         if not os.path.exists("img"):
@@ -1032,10 +1049,12 @@ class CLASSIX:
             else: # explain(index1, index2)
                 if isinstance(index2, int):
                     object2 = self.x_pca[index2] # self.data has been normalized
+                    agg_label2 = self.agg_labels[index2] # get the group index for object2
                 else:
                     object2 = (index2 - self._mu) / self._scl # allow for out-sample data
                     if self.data.shape[1] >= 2:
                         object2 = self.pca.transform(object2)
+                    agg_label2 = np.argmin(np.linalg.norm(self.s_pca - object2, axis=1, ord=2)) # get the group index for object2
 
                 if replace_name != None:
                     if isinstance(replace_name, list):
@@ -1047,9 +1066,6 @@ class CLASSIX:
                 else:
                     index1_name = index1
                     index2_name = index2
-                        
-                agg_label1 = np.argmin(np.linalg.norm(self.s_pca - object1, axis=1, ord=2)) # get the group index for object1
-                agg_label2 = np.argmin(np.linalg.norm(self.s_pca - object2, axis=1, ord=2)) # get the group index for object2
 
                 cluster_label1, cluster_label2 = self.label_change[agg_label1], self.label_change[agg_label2]
 
@@ -1061,18 +1077,104 @@ class CLASSIX:
                         connected_groups["object 1"] = np.array(subs, dtype=int)
                     elif agg_label2 in subs:
                         connected_groups["object 2"] = np.array(subs, dtype=int)
+               
+                
+                if agg_label1 == agg_label2: # when ind1 & ind2 are in the same group
+                    sp_str = np.array2string(self.splist[agg_label1, 3:], separator=',')
+                    print("The data points %(index1)s and %(index2)s are in the same group %(agg_id)i, hence were merged into the same cluster #%(m_c)i"%{
+                        "index1":index1, "index2":index2, "agg_id":agg_label1, "m_c":cluster_label1}
+                    )
+                    connected_paths = [agg_label1]
+                else:
+                    sp1_str = self.splist[agg_label1, 3:]
+                    sp2_str = self.splist[agg_label2, 3:]
+                    # print("""The two objects are assigned to different groups through aggregation.""")
 
+                    # print(
+                    # """The first object is aggregated to group %(agg_id)i associated with starting point %(sp)s with respect to radius %(radius)s, then assigned to cluster label %(m_c)i.""" % {
+                    #     "agg_id":agg_label1, "sp":sp1_str, "radius":self.radius, "m_c":cluster_label1
+                    # })
+
+                    # print(
+                    # """The second object is aggregated to group %(agg_id)i associated with starting point %(sp)s with respect to radius %(radius)s, then assigned to cluster label %(m_c)i.""" % {
+                    #     "agg_id":agg_label2, "sp":sp2_str, "radius":self.radius, "m_c":cluster_label2
+                    # })
+                    
+                    if self.connected_pairs is None:
+                        distm = pairwise_distances(self.splist[:,3:], Y=None, metric='euclidean', n_jobs=n_jobs)
+                        distm = (distm <= radius*scale).astype(int)
+                        self.connected_pairs = return_csr_matrix_indices(csr_matrix(distm)).tolist() # list
+                        
+                    if cluster_label1 == cluster_label2: # when ind1 & ind2 are in the same cluster but diff group
+                        # deprecated (24/07/2021)  from scipy.sparse.csgraph import shortest_path -> Dijkstra’s algorithm 
+                        # path_graph = pairs_to_graph(self.connected_pairs, N=self.splist.shape[0], sparse=True)
+                        ## apply Dijkstra’s algorithm with Fibonacci heaps for shortest path finding
+                        # dist_matrix, predecessors = shortest_path(csgraph=path_graph, method="D", directed=False, return_predecessors=True) 
+                        # connected_paths = get_shortest_path(predecessors, agg_label1, agg_label2)
+                        connected_paths = find_shortest_path(agg_label1,
+                                                             self.connected_pairs,
+                                                             self.splist.shape[0],
+                                                             agg_label2
+                        )
+                        
+                        # -- Deprecated
+                        # if agg_label1 != 0:
+                        #     connected_paths_vis = " -> ".join([str(group) for group in connected_paths]) 
+                        # else:
+                        #     connected_paths_vis = " -> ".join(["0"] + [str(group) for group in connected_paths]) 
+                        
+                        connected_paths_vis = " <-> ".join([str(group) for group in connected_paths]) 
+                        print(
+                        """The data point %(index1)s is in group %(agg_id1)s and the data point %(index2)s is in group %(agg_id2)s, """
+                            """\nboth of which were merged into cluster #%(cluster)i. """% {
+                            "index1":index1, "index2":index2, "cluster":cluster_label1, "agg_id1":agg_label1, "agg_id2":agg_label2}
+                        )
+                        
+                        print("""These two groups are connected via groups %(connected)s.""" % {
+                               "connected":connected_paths_vis}
+                        )
+                    else: 
+                        # alternative expression 1
+                        # print("The two objects are allocated to two different cluster.")
+                        # print(
+                        # """The first object is allocated to the cluster %(c_id1)s via %(connected1)s with cluster center %(cen1)s
+                        # while the second object is allocated to the cluster %(c_id2)s via %(connected2)s with cluster center %(cen2)s""" % {
+                        # "c_id1":cluster_label1, "connected1": connected_groups["object 1"], "cen1":cen1_str,
+                        # "c_id2":cluster_label2, "connected2": connected_groups["object 2"], "cen1":cen2_str
+                        #     }
+                        # )
+                        # print("The two objects are allocated to two different cluster.")
+                        
+                        # alternative expression 2
+                        # print("""The data point %(index1)s is in group %(agg_id1)i, which has been merged into cluster %(c_id1)s via connected groups %(connected1)s.""" % {
+                        #     "index1":index1, "agg_id1":agg_label1, "c_id1":cluster_label1, "connected1": connected_groups["object 1"]
+                        # })
+
+                        # print("""\nThe data point %(index2)s is in group %(agg_id2)i, which has been merged into cluster %(c_id2)s via connected groups %(connected2)s.""" % {
+                        #     "index2":index2, "agg_id2":agg_label2, "c_id2":cluster_label2, "connected2": connected_groups["object 2"]
+                        # })
+                        connected_paths = []
+                        print("""The data point %(index1)s is in group %(agg_id1)i, which has been merged into cluster %(c_id1)s.""" % {
+                            "index1":index1, "agg_id1":agg_label1, "c_id1":cluster_label1})
+
+                        print("""The data point %(index2)s is in group %(agg_id2)i, which has been merged into cluster %(c_id2)s.""" % {
+                            "index2":index2, "agg_id2":agg_label2, "c_id2":cluster_label2})   
+                        
+                        print("""There is no path of overlapping groups between these clusters.""")
+                        
                 if plot == True:
                     plt.style.use(style=figstyle)
                     fig, ax = plt.subplots(figsize=figsize)
                     plt.rcParams['axes.facecolor'] = 'whitesmoke'
                     # select indices
+                    union_ind = np.where((self.sp_info.Cluster == cluster_label1) | (self.sp_info.Cluster == cluster_label2))[0]
                     x_pca1 = self.x_pca[self.labels_ == cluster_label1]
                     x_pca2 = self.x_pca[self.labels_ == cluster_label2]
-                    s_pca = self.s_pca[(self.sp_info.Cluster == cluster_label1) | (self.sp_info.Cluster == cluster_label2)]
+                    s_pca = self.s_pca[union_ind]
+                    # self.s_pca[]
                     
-                    ax.scatter(x_pca1[:, 0], x_pca1[:, 1], marker="*", c=self.cluster_color[cluster_label1])
-                    ax.scatter(x_pca2[:, 0], x_pca2[:, 1], marker="*", c=self.cluster_color[cluster_label2])
+                    ax.scatter(x_pca1[:, 0], x_pca1[:, 1], marker=".", c=self.cluster_color[cluster_label1])
+                    ax.scatter(x_pca2[:, 0], x_pca2[:, 1], marker=".", c=self.cluster_color[cluster_label2])
                     
                     # ax.scatter(x_pca1[:, 0], x_pca1[:, 1],
                     #             marker="*", c='lightgreen')
@@ -1080,20 +1182,32 @@ class CLASSIX:
                     #             marker="*", c='lightblue')
                     
                     ax.scatter(s_pca[:,0], s_pca[:,1], marker="p")
-
-                    ax.scatter(object1[0],  object1[1], marker="*", s=ind_marker_size)
-                    ax.text(object1[0], object1[1], s=str(index1_name), color=ind_color)
-
-                    ax.scatter(object2[0],  object2[1], marker="*", s=ind_marker_size)
-                    ax.text(object2[0], object2[1], s=str(index2_name), color=ind_color)
+                    ax.scatter(object1[0], object1[1], marker="*", s=ind_marker_size)
+                    ax.scatter(object2[0], object2[1], marker="*", s=ind_marker_size)
+                    
+                    if dp_fontsize is None:
+                        ax.text(object1[0], object1[1], s=str(index1_name), bbox=dp_bbox, color=ind_color)
+                        ax.text(object2[0], object2[1], s=str(index2_name), bbox=dp_bbox, color=ind_color)
+                    else:
+                        ax.text(object1[0], object1[1], s=str(index1_name), fontsize=dp_fontsize, bbox=dp_bbox, color=ind_color)
+                        ax.text(object2[0], object2[1], s=str(index2_name), fontsize=dp_fontsize, bbox=dp_bbox, color=ind_color)
 
                     for i in range(s_pca.shape[0]):
-                        ax.add_patch(plt.Circle((s_pca[i, 0], s_pca[i, 1]), self.radius, fill=False, color=color, alpha=alpha, lw=cline_width, clip_on=False))
+                        if union_ind[i] in connected_paths:
+                            ax.add_patch(plt.Circle((s_pca[i, 0], s_pca[i, 1]), self.radius, fill=False, color=connect_color, alpha=alpha, lw=cline_width, clip_on=False))
+                        else:
+                            ax.add_patch(plt.Circle((s_pca[i, 0], s_pca[i, 1]), self.radius, fill=False, color=color, alpha=alpha, lw=cline_width, clip_on=False))
                         ax.set_aspect('equal', adjustable='datalim')
-                        ax.text(s_pca[i, 0], s_pca[i, 1], 
-                            s=self.sp_info.Group[(self.sp_info.Cluster == cluster_label1) | (self.sp_info.Cluster == cluster_label2)].values[i].astype(int).astype(str),
-                            bbox=bbox)
-                        
+                        if sp_fontsize is None:
+                            ax.text(s_pca[i, 0], s_pca[i, 1], 
+                                    s=self.sp_info.Group[(self.sp_info.Cluster == cluster_label1) | (self.sp_info.Cluster == cluster_label2)].values[i].astype(int).astype(str),
+                                    bbox=sp_bbox
+                            )
+                        else:
+                            ax.text(s_pca[i, 0], s_pca[i, 1], 
+                                    s=self.sp_info.Group[union_ind].values[i].astype(int).astype(str),
+                                    fontsize=sp_fontsize, bbox=sp_bbox
+                            )
                     ax.axis('off') # the axis here may not be consistent, so hide.
                     ax.plot()
                     if savefig:
@@ -1103,7 +1217,7 @@ class CLASSIX:
                             plt.savefig('img/' + str(figname) + str(index1) + '_' + str(index2) +'.pdf', bbox_inches='tight')
                         else:
                             plt.savefig('img/' + str(figname) + str(index1) + '_' + str(index2) +'.png', bbox_inches='tight')
-                        print("successfully save")
+                        print("image successfully save.")
                     plt.show()
                     
                     # deprecated (24/07/2021):
@@ -1131,7 +1245,7 @@ class CLASSIX:
                     #         ax.set_aspect('equal', adjustable='datalim')
                     #         ax.text(s_pca[i, 0], s_pca[i, 1], 
                     #             s=self.sp_info.Group[(self.sp_info.Cluster == cluster_label1) | (self.sp_info.Cluster == cluster_label2)].values[i].astype(int).astype(str),
-                    #             bbox=bbox)
+                    #             bbox=sp_bbox)
                     #     
                     #     ax.plot()
                     #     if savefig:
@@ -1164,7 +1278,7 @@ class CLASSIX:
                     #         ax.set_aspect('equal', adjustable='datalim')
                     #         ax.text(s_pca[i, 0], s_pca[i, 1],
                     #             s=self.sp_info.Group[(self.sp_info.Cluster == cluster_label1) | (self.sp_info.Cluster == cluster_label2)].values[i].astype(int).astype(str), 
-                    #             bbox=bbox)
+                    #             bbox=sp_bbox)
                     #     
                     #     ax.plot()
                     #     if savefig:
@@ -1177,88 +1291,7 @@ class CLASSIX:
                     # else:
                     #     print("This function is restricted to multidimensional (dimension greater than or equal to 2) data.")
                         
-                if agg_label1 == agg_label2: # when ind1 & ind2 are in the same group
-                    sp_str = np.array2string(self.splist[agg_label1, 3:], separator=',')
-                    print("The data points %(index1)s and %(index2)s are in the same group %(agg_id)i, hence were merged into the same cluster #%(m_c)i"%{
-                        "index1":index1, "index2":index2, "agg_id":agg_label1, "m_c":cluster_label1}
-                    )
 
-                else:
-                    sp1_str = self.splist[agg_label1, 3:]
-                    sp2_str = self.splist[agg_label2, 3:]
-                    # print("""The two objects are assigned to different groups through aggregation.""")
-
-                    # print(
-                    # """The first object is aggregated to group %(agg_id)i associated with starting point %(sp)s with respect to radius %(radius)s, then assigned to cluster label %(m_c)i.""" % {
-                    #     "agg_id":agg_label1, "sp":sp1_str, "radius":self.radius, "m_c":cluster_label1
-                    # })
-
-                    # print(
-                    # """The second object is aggregated to group %(agg_id)i associated with starting point %(sp)s with respect to radius %(radius)s, then assigned to cluster label %(m_c)i.""" % {
-                    #     "agg_id":agg_label2, "sp":sp2_str, "radius":self.radius, "m_c":cluster_label2
-                    # })
-                    
-                    if self.connected_pairs is None:
-                        distm = pairwise_distances(self.splist[:,3:], Y=None, metric='euclidean', n_jobs=n_jobs)
-                        distm = (distm <= radius*scale).astype(int)
-                        self.connected_pairs = return_csr_matrix_indices(csr_matrix(distm)).tolist() # list
-                        
-                    if cluster_label1 == cluster_label2: # when ind1 & ind2 are in the same cluster but diff group
-                        # deprecated (24/07/2021)  from scipy.sparse.csgraph import shortest_path -> Dijkstra’s algorithm 
-                        # path_graph = pairs_to_graph(self.connected_pairs, N=self.splist.shape[0], sparse=True)
-                        ## apply Dijkstra’s algorithm with Fibonacci heaps for shortest path finding
-                        # dist_matrix, predecessors = shortest_path(csgraph=path_graph, method="D", directed=False, return_predecessors=True) 
-                        # connected_paths = get_shortest_path(predecessors, agg_label1, agg_label2)
-                        connected_paths = find_shortest_path(agg_label1,
-                                                             self.connected_pairs,
-                                                             self.splist.shape[0],
-                                                             agg_label2)
-                        
-                        # -- Deprecated
-                        # if agg_label1 != 0:
-                        #     connected_paths_vis = " -> ".join([str(group) for group in connected_paths]) 
-                        # else:
-                        #     connected_paths_vis = " -> ".join(["0"] + [str(group) for group in connected_paths]) 
-                        
-                        connected_paths_vis = " <-> ".join([str(group) for group in connected_paths]) 
-                        print(
-                        """The data point %(index1)s is in group %(agg_id1)s and the data point %(index2)s is in group %(agg_id2)s, """
-                            """\nboth of which were merged into cluster #%(cluster)i. """% {
-                            "index1":index1, "index2":index2, "cluster":cluster_label1, "agg_id1":agg_label1, "agg_id2":agg_label2}
-                        )
-                        
-                       
-                        print("""These two groups are connected via groups %(connected)s.""" % {
-                               "connected":connected_paths_vis}
-                        )
-                    else: 
-                        # alternative expression 1
-                        # print("The two objects are allocated to two different cluster.")
-                        # print(
-                        # """The first object is allocated to the cluster %(c_id1)s via %(connected1)s with cluster center %(cen1)s
-                        # while the second object is allocated to the cluster %(c_id2)s via %(connected2)s with cluster center %(cen2)s""" % {
-                        # "c_id1":cluster_label1, "connected1": connected_groups["object 1"], "cen1":cen1_str,
-                        # "c_id2":cluster_label2, "connected2": connected_groups["object 2"], "cen1":cen2_str
-                        #     }
-                        # )
-                        # print("The two objects are allocated to two different cluster.")
-                        
-                        # alternative expression 2
-                        # print("""The data point %(index1)s is in group %(agg_id1)i, which has been merged into cluster %(c_id1)s via connected groups %(connected1)s.""" % {
-                        #     "index1":index1, "agg_id1":agg_label1, "c_id1":cluster_label1, "connected1": connected_groups["object 1"]
-                        # })
-
-                        # print("""\nThe data point %(index2)s is in group %(agg_id2)i, which has been merged into cluster %(c_id2)s via connected groups %(connected2)s.""" % {
-                        #     "index2":index2, "agg_id2":agg_label2, "c_id2":cluster_label2, "connected2": connected_groups["object 2"]
-                        # })
-
-                        print("""The data point %(index1)s is in group %(agg_id1)i, which has been merged into cluster %(c_id1)s.""" % {
-                            "index1":index1, "agg_id1":agg_label1, "c_id1":cluster_label1})
-
-                        print("""The data point %(index2)s is in group %(agg_id2)i, which has been merged into cluster %(c_id2)s.""" % {
-                            "index2":index2, "agg_id2":agg_label2, "c_id2":cluster_label2})   
-                        
-                        print("""There is no path of overlapping groups between these clusters.""")
                         
                 # if showsplist:
                 #     print(dash_line)
@@ -1268,12 +1301,11 @@ class CLASSIX:
                 #     print(spdisplay.to_string(justify='center', index=False, max_colwidth=max_colwidth))
                 #     print(dash_line)
                 
-
         return 
     
     
     
-    def explain_viz(self, figsize=(12, 8), figstyle="ggplot", savefig=False, bbox={'facecolor': 'tomato', 'alpha': 0.2, 'pad': 2}, fmt='pdf'):
+    def explain_viz(self, figsize=(12, 8), figstyle="ggplot", savefig=False, fontsize=None, bbox={'facecolor': 'tomato', 'alpha': 0.3, 'pad': 2}, axis='off', fmt='pdf'):
         """
             Visualize the starting point and data points
         """
@@ -1292,8 +1324,13 @@ class CLASSIX:
                 plt.scatter(x_pca_part[:,0], x_pca_part[:,1], marker="*", c=self.cluster_color[i])
                 # plt.scatter(x_pca_part[:,0], x_pca_part[:,1], marker="*", c='#%06X' % np.random.randint(0, 0xFFFFFF))
                 for j in range(self.s_pca.shape[0]):
-                    plt.text(self.s_pca[j, 0], self.s_pca[j, 1], str(j), bbox=bbox)
-                plt.axis('off') # the axis here may not be consistent, so hide.
+                    if fontsize is None:
+                        plt.text(self.s_pca[j, 0], self.s_pca[j, 1], str(j), bbox=bbox)
+                    else:
+                        plt.text(self.s_pca[j, 0], self.s_pca[j, 1], str(j), fontsize=fontsize, bbox=bbox)
+            plt.axis(axis) # the axis here may not be consistent, so hide.
+            plt.xlim([np.min(self.s_pca[:,0])-0.1, np.max(self.s_pca[:,0])+0.1])
+            plt.ylim([np.min(self.s_pca[:,1])-0.1, np.max(self.s_pca[:,1])+0.1])
         else:
             print("Visualization is restricted to multidimensional (dimension greater than or equal to 2) data.")
             
